/*
*/

#include "globals.h"
#include <avr/eeprom.h>

// variables that are stored in eeprom
uint8_t g_function_id								__attribute__ ((section (".noinit"))) ;
uint8_t g_function_data							__attribute__ ((section (".noinit"))) ;
uint8_t g_chain_length							__attribute__ ((section (".noinit"))) ;
uint8_t g_chain_pos									__attribute__ ((section (".noinit"))) ;
uint16_t g_meter_min								__attribute__ ((section (".noinit"))) ;
uint16_t g_meter_max								__attribute__ ((section (".noinit"))) ;
uint8_t g_tickdelay									__attribute__ ((section (".noinit"))) ;
uint8_t g_transmit									__attribute__ ((section (".noinit"))) ;

// derived variables
uint8_t g_myfirstled			__attribute__ ((section (".noinit"))) ;
uint8_t g_chain_ledlength	__attribute__ ((section (".noinit"))) ;

// define how many bytes of wear levelling will be applied to function_data
#define WEARLEVELSIZE_function_data 10

typedef struct {
	uint8_t DUMMY ; // don't store data in position 0. It's unlucky.
	uint8_t function_id ;
	uint8_t function_data[ WEARLEVELSIZE_function_data ] ; //the first non 0xff value will be used.
	uint8_t chain_length ;
	uint8_t chain_pos ;
	uint16_t meter_min ;
	uint16_t meter_max ;
	uint8_t tickdelay ;
	uint8_t transmit ;
	uint8_t pov_bitmap[128] ;
	uint8_t sweep_userpattern_data[130] ;
	} EData_t ;
EData_t EEMEM EData = {
	0xff, // DUMMY
	FUNCTION_SWEEP,
	{0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, 0 },
	1,
	1,
	0,
	1023,
	40, // tickdelay of 40 means a 4ms 'tick' at 'medium' speed.
	0, // no transmit by default
	{
		0xff,0x7e,0x3c,0x1c,0x08,0x81,0x81,0x81,0x81,0x00,0x3c,0x42,0x99,0xa5,0xa5,0x99,0x42,0x3c,0x00,0xff,0x81,0xbd,0xa5,0xa5,0xbd,0x81,0xff,0x00,0xff,0x81,0x42,0x24,
		0x18,0x00,0x00,0xff,0xff,0x18,0x18,0xff,0xff,0x00,0x00,0xff,0xff,0xdb,0xdb,0xc3,0x00,0x00,0xff,0xff,0x03,0x03,0x03,0x00,0x00,0xff,0xff,0x03,0x03,0x03,0x00,0x00,
		0x7e,0xff,0xe7,0xc3,0xe7,0xff,0x7e,0x00,0x00,0x70,0xfd,0xfd,0xfd,0x70,0x00,0x00,0x00,0x00,0x00,0x00,0xaa,0x55,0xaa,0x55,0xaa,0x55,0xaa,0x55,0xaa,0x00,0x33,0x33,
		0xcc,0xcc,0x33,0x33,0xcc,0xcc,0x33,0x33,0x00,0xf0,0xf0,0xf0,0xf0,0x0f,0x0f,0x0f,0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
	},
	{
		128, // length
		1, // delay (in ticks) between bytes
		0x80,0x80,0x80,0x80,0x80,0x80,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
		0x40,0x40,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x10,0x10,0x10,0x10,0x10,0x10,
		0x10,0x08,0x08,0x08,0x08,0x08,0x08,0x04,0x04,0x04,0x04,0x04,0x04,0x04,0x04,0x02,
		0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x01,0x01,0x01,0x01,0x01,
		0x01,0x01,0x01,0x01,0x01,0x01,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,
		0x02,0x02,0x04,0x04,0x04,0x04,0x04,0x04,0x04,0x04,0x08,0x08,0x08,0x08,0x08,0x08,
		0x08,0x10,0x10,0x10,0x10,0x10,0x10,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x40,
		0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x80,0x80,0x80,0x80,0x80
	}
} ;

uint8_t eeprom_read_wearlevelled_byte( void *buffer, uint8_t bufsize ){
// NOTE: it is assumed that the value 0xff will never be stored.
// However; 0xff will be returned if there are no valid values in the buffer.
// The buffer is filled from top to bottom, so when reading the buffer
// from bottom to top, the most recent value is read.
	for( uint8_t index = 0 ; index < bufsize ; index++ ){
		uint8_t value = eeprom_read_byte( buffer + index );
		if (value != 0xff) return( value ) ;
	}
	return 0xff ;
}

void eeprom_update_wearlevelled_byte( void *buffer, uint8_t bufsize, uint8_t value ){
// NOTE: it is assumed that the value 0xff is never valid.

// First check to see if the most recent byte is the same value.
// if it is, there's no need to write anything.
	if (eeprom_read_wearlevelled_byte( buffer, bufsize ) == value) return ;

// It's time to write a byte...
// First of all, check that there is at least one spare empty place to
// write a value...

	if ( eeprom_read_byte( buffer ) == 0xff ){
		// there's at least one spare place to write a new value.
		// Find it and write it...
		void *b = buffer ;
		b += bufsize - 1 ;
		for( uint8_t count = 0 ; count < bufsize ; count++ ){
			if ( eeprom_read_byte( b ) == 0xff ){
				// write the byte. It doesn't need to be cleared first.
				eeprom_busy_wait();
				EEAR = (uint16_t)b ;
				EEDR = value ;
				EECR = (1<<EEPM1)|(0<<EEPM0)|(1<<EEMPE) ;
				EECR = (1<<EEPM1)|(0<<EEPM0)|(1<<EEMPE)|(1<<EEPE) ;
				// the byte is now being written.
				EEAR = 0 ; // just in case.
				return ; // all done!
			}
			b-- ;
		}
	}
	
	// there was no space in the buffer...
	// erase all the bytes in the buffer (except the last...)
	for( uint8_t index = 0 ; index < bufsize-1 ; index++ ){
		eeprom_busy_wait();
		EEAR = (uint16_t)(buffer + index) ;
		EECR = (0<<EEPM1)|(1<<EEPM0)|(1<<EEMPE) ;
		EECR = (0<<EEPM1)|(1<<EEPM0)|(1<<EEMPE)|(1<<EEPE) ;
	}
	// now erase _and_ write the last byte in the buffer...
	eeprom_busy_wait();
	EEAR = (uint16_t)(buffer + bufsize - 1) ;
	EEDR = value ;
	EECR = (0<<EEPM1)|(0<<EEPM0)|(1<<EEMPE) ;
	EECR = (0<<EEPM1)|(0<<EEPM0)|(1<<EEMPE)|(1<<EEPE) ;
	EEAR = 0 ;

}

void eeprom_restore_state( void ){
	// pull values for global variables from eeprom.
	g_function_id		= eeprom_read_byte( &(EData.function_id) ) ;
	g_function_data	= eeprom_read_wearlevelled_byte( &(EData.function_data), WEARLEVELSIZE_function_data ) ;
	g_chain_length	= eeprom_read_byte( &(EData.chain_length) ) ;
	g_chain_pos			= eeprom_read_byte( &(EData.chain_pos) ) ;
	g_meter_min			= eeprom_read_word( &(EData.meter_min) ) ;
	g_meter_max			= eeprom_read_word( &(EData.meter_max) ) ;
	g_tickdelay			= eeprom_read_byte( &(EData.tickdelay) ) ;
	g_transmit			= eeprom_read_byte( &(EData.transmit) ) ;
	
	// do a quick sanity check for the values;
	// chain length must be between 1 and 4 inclusive
	if (g_chain_length == 0) g_chain_length = 1 ;
	if (g_chain_length > 4) g_chain_length = 4 ;
	// chain position must be within the length of the chain(!)
	if (g_chain_pos == 0) g_chain_pos = 1 ;
	if (g_chain_pos > g_chain_length) g_chain_pos = g_chain_length ;
	
	// update derived values ;
	g_myfirstled = (g_chain_pos - 1) * 8 ;
	g_chain_ledlength = g_chain_length * 8 ;
}

void eeprom_save_state( void ){
	// write values from global variables into eeprom.
	// Use the 'update' variants, so they're only written if the value's changed.
	eeprom_update_byte( &(EData.function_id), g_function_id ) ;
	eeprom_update_wearlevelled_byte( &(EData.function_data) , WEARLEVELSIZE_function_data, g_function_data ) ;
	eeprom_update_byte( &(EData.chain_length), g_chain_length ) ;
	eeprom_update_byte( &(EData.chain_pos), g_chain_pos ) ;
	eeprom_update_word( &(EData.meter_min), g_meter_min ) ;
	eeprom_update_word( &(EData.meter_max), g_meter_max ) ;
	eeprom_update_byte( &(EData.tickdelay), g_tickdelay ) ;
	eeprom_update_byte( &(EData.transmit), g_transmit ) ;
}

//inline uint8_t *eeprom_get_povbitmap(){ return EData.pov_bitmap ; }
inline uint8_t *eeprom_get_sweepuserpattern(){ return EData.sweep_userpattern_data ; }
