/*
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "globals.h"
#include "led.h"
#include "receive.h"
#include "random.h"

#ifndef TIMSK
#define TIMSK TIMSK0
#endif

uint8_t PROGMEM f_ledmap[] = {
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x01,0x01,0x02,0x02,0x02,0x03,0x03,
	0x04,0x04,0x04,0x05,0x05,0x06,0x06,0x06,0x07,0x07,0x08,0x08,0x09,0x09,0x0A,0x0A,
	0x0B,0x0B,0x0C,0x0C,0x0D,0x0E,0x0E,0x0F,0x0F,0x10,0x11,0x11,0x12,0x12,0x13,0x14,
	0x14,0x15,0x16,0x16,0x17,0x18,0x18,0x19,0x1A,0x1A,0x1B,0x1C,0x1D,0x1D,0x1E,0x1F,
	0x20,0x20,0x21,0x22,0x23,0x23,0x24,0x25,0x26,0x27,0x27,0x28,0x29,0x2A,0x2B,0x2B,
	0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x39,
	0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,
	0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,
	0x5A,0x5B,0x5C,0x5D,0x5E,0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6B,
	0x6C,0x6D,0x6E,0x6F,0x70,0x71,0x73,0x74,0x75,0x76,0x77,0x78,0x7A,0x7B,0x7C,0x7D,
	0x7E,0x7F,0x81,0x82,0x83,0x84,0x85,0x87,0x88,0x89,0x8A,0x8C,0x8D,0x8E,0x8F,0x90,
	0x92,0x93,0x94,0x95,0x97,0x98,0x99,0x9B,0x9C,0x9D,0x9E,0xA0,0xA1,0xA2,0xA4,0xA5,
	0xA6,0xA7,0xA9,0xAA,0xAB,0xAD,0xAE,0xAF,0xB1,0xB2,0xB3,0xB5,0xB6,0xB7,0xB9,0xBA,
	0xBB,0xBD,0xBE,0xBF,0xC1,0xC2,0xC4,0xC5,0xC6,0xC8,0xC9,0xCA,0xCC,0xCD,0xCF,0xD0,
	0xD1,0xD3,0xD4,0xD6,0xD7,0xD9,0xDA,0xDB,0xDD,0xDE,0xE0,0xE1,0xE3,0xE4,0xE5,0xE7,
	0xE8,0xEA,0xEB,0xED,0xEE,0xF0,0xF1,0xF3,0xF4,0xF6,0xF7,0xF9,0xFA,0xFC,0xFD,0xFF
} ;

typedef void effect_fn( void ) ;

/* renderers */
effect_fn mapstraight ;	// direct mapping
//effect_fn mapavg ;		// maps using an average of surrounding output values
effect_fn mapsparkle ;
effect_fn mapdrift ;	// slowly drifts the output linearly towards the pattern value
effect_fn mapexp ;		// drags the output towards the pattern value in an exponential manner (i.e. the closer to the value, the slower the movement towards it)
//effect_fn mapdriftdown ;
//effect_fn mapflicker ;	// flickers the output with random off cycles.


volatile uint8_t g_tick __attribute__ ((section (".noinit"))) ;
uint8_t g_led_target[8] __attribute__ ((section (".noinit"))) ;
uint8_t g_led_display[8] __attribute__ ((section (".noinit")));

static StyleID_t m_style = STYLE_Exponential ;

void _decode_command( uint16_t cmd ) ;

void inline led_seteffect( StyleID_t e ){
	m_style = e ;
}

// channel 0 is LEDs 1,3,5,7
// channel 1 is LEDs 2,4,6,8
#define CHANNELCOUNT 2
#define BITCOUNT 8
static uint8_t g_bitslice[BITCOUNT * CHANNELCOUNT] __attribute__ ((section (".noinit")));
// Note that g_bitslice must not have PB2 set.

inline void led_init( void )
{
	for( uint8_t i = 0 ; i < 8 ; i++ ){ g_led_display[i] = 0; g_led_target[i] = 0; }
	for( uint8_t i = 0 ; i < (BITCOUNT * CHANNELCOUNT) ; i++ ){ g_bitslice[i] = 0; }
	DDRB &= SENSE_bm ; // set every pin (excluding sense pin) as input
	PORTB &= SENSE_bm ; // clear all bits (excluding sense pin).
#ifdef DEBUG
	DDRB |= (1<<PB0) ; // set PB0 as output
	PORTB &= ~(1<<PB0) ; // clear PB0
#endif
	TCCR0A	= (1<<WGM01)|(0<<WGM00) ; // set the timer to CTC mode.
	TCCR0B	= (0<<WGM02)|(0<<CS00)|(1<<CS01)|(1<<CS00) ; // and a /64 prescaler.
	OCR0A		= 0b10000000 ; // initial delay (essentially it will fire when the first SEI is run)
	TCNT0		= 0 ; // start the timer now.
	TIMSK	= (1 << OCIE0A) ; // Enable the Compare Match interrupt
	g_tick = 0 ;
}

void led_encode_timeslice(){
// this method takes the g_led_display array of 8 bytes, and calculates
// the values to but put onto the I/O lines for each of the timeslices.

// LED1 = PB4 high
// LED2 = PB4 low
// LED3 = PB3 high
// LED4 = PB3 low
// LED5 = PB1 high
// LED6 = PB1 low
// LED7 = PB0 high
// LED8 = PB0 low
//
//
	uint8_t bitvalue = 0b00000001 ;
	uint8_t index = 0 ;
	uint8_t mapped[8] ;
	for( uint8_t i = 0 ; i < 8 ; i++ ){
		mapped[i] = pgm_read_byte( f_ledmap + g_led_display[i] ) ;
	}
	
	for ( uint8_t bitpos = 0 ; bitpos < 8 ; bitpos++ ){
		uint8_t portbits_chan1 = 0 ;
		uint8_t portbits_chan2 = 0 ;
		uint8_t pinmask = 0b00010000 ;
		for( uint8_t i = 0 ; i < 8 ; i += 2 ){
			
			if (mapped[i] & bitvalue) portbits_chan2 |= pinmask ;
			if (mapped[i+1] & bitvalue) portbits_chan1 |= pinmask ;
			pinmask >>= 1 ;
			if (i == 2) pinmask >>= 1 ; 
		}
#ifdef DEBUG
		// ensure that PB0 is left alone.
		portbits_chan1 &= ~(1<<PB0) ;
		portbits_chan2 &= ~(1<<PB0) ;
#endif
		g_bitslice[ index++ ] = portbits_chan1 ;
		g_bitslice[ index++ ] = portbits_chan2 ;
		bitvalue <<= 1 ;
	}
}

ISR( TIM0_COMPA_vect )
{
static uint8_t g_index = 0 ;
	if (g_index > 15){
		// turn off LEDs whilst the (longish duration) updates are done.
#ifdef DEBUG
		DDRB &= (SENSE_bm | (1<<PB0)) ; // set every pin (excluding sense pin and PB0) as input
		PORTB &= (SENSE_bm | (1<<PB0)) ; // clear all bits (excluding sense pin and PB0).
#else
		DDRB &= SENSE_bm ; // set every pin (excluding sense pin) as input
		PORTB &= SENSE_bm ; // clear all bits (excluding sense pin).
#endif
		g_index = 0 ;
		g_tick = 1 ;
		OCR0A = 0b10000000 ;
		
		// disable this interrupt...
		TIMSK &= ~(1<<OCIE0A) ;
		// enable other interrupts...
		// (e.g. allow external_interrupt ISRs to receive bit data)
		sei();
		
		// check to see if we've received any commands...
		while( rx_ready() ){
			_decode_command( rx_getdata() ) ;
		}
		
		if (m_style == STYLE_Exponential) mapexp();
		if (m_style == STYLE_Direct) mapstraight();
		if (m_style == STYLE_Sparkle) mapsparkle();
		if (m_style == STYLE_Drift) mapdrift() ;
		led_encode_timeslice();
		
		// now we've finished the long duration processing,
		// switch back to not allowing interrupts
		// (they'll ultimately be turned on again when this ISR is exited)
		cli() ;
		
		// and now re-start timers and re-enable this interrupt...
		GTCCR |= (1<<PSR0) ; // reset the prescaler
		TCNT0 = 0 ;
		TIFR = (1<<OCF0A) ; // clear any pending interrupts.
		TIMSK |= (1<<OCIE0A) ; // re-enable _this_ interrupt.
	}
	uint8_t bits = g_bitslice[ g_index ] ;
#ifdef DEBUG
		DDRB &= (SENSE_bm | (1<<PB0)) ; // set every pin (excluding sense pin and PB0) as input
		PORTB &= (SENSE_bm | (1<<PB0)) ; // clear all bits (excluding sense pin and PB0).
#else
		DDRB &= SENSE_bm ; // set every pin (excluding sense pin) as input
		PORTB &= SENSE_bm ; // clear all bits (excluding sense pin).
#endif

	// set any bits required - note, SENSE pin should be left alone.
	DDRB |= bits ; // note that the sense pin will always be 0 in 'bits'.

	if (g_index & 0b1){
		PORTB |= bits ;
	} else {
		uint8_t temp = OCR0A << 1 ;
		if (temp==0){ temp = 0b00000001 ;}
		OCR0A = temp ;
	}
	g_index++ ;
}

#define LEDCOUNT 8

void mapstraight()
{
	for( uint8_t ledindex = 0 ; ledindex < LEDCOUNT ; ledindex++ ){
		g_led_display[ ledindex ] = g_led_target[ledindex] ;
	}
}

void mapsparkle(){
static uint8_t s_count = 1 ;
static uint8_t s_mask = 0 ;

	if (s_count-- == 0){
		 s_mask = random_getbyte() ;
		 s_count = 10  ;
	}

	uint8_t mask = s_mask ;
	for( uint8_t ledindex = 0 ; ledindex < LEDCOUNT ; ledindex++ ){
		uint8_t value = (mask & 0b1) ? 255 : 0 ;
		mask >>= 1 ;
		g_led_display[ ledindex ] = g_led_target[ ledindex ] & 0b10000000 ? value : 0 ;
	}
}

void mapdrift(){
	for( uint8_t ledindex = 0 ; ledindex < LEDCOUNT ; ledindex++ )
	{
		uint8_t i = g_led_display[ ledindex ] ;
		uint8_t t = g_led_target[ ledindex ] ;
		if ( i < t ){
			uint8_t delta = t - i ;
			if (delta > 8) delta = 8 ; // step by 4
			g_led_display[ ledindex ] = g_led_display[ ledindex ] + delta ;
		}
		if ( i > t ){
			uint8_t delta =  i - t ;
			if (delta > 1) delta = 1 ; // step by 4
			if (delta == 0) delta = 1 ;
			g_led_display[ ledindex ] = g_led_display[ ledindex ] - delta ;
		}
	}
}

void mapexp()
{
	for( uint8_t ledindex = 0 ; ledindex < LEDCOUNT ; ledindex++ )
	{
		uint8_t i = g_led_display[ ledindex ] ;
		uint8_t t = g_led_target[ ledindex ] ;
		if ( i < t ){
			uint8_t delta = t - i ;
			delta = delta >> 4 ;// divide by 8
			if (delta == 0) delta = 1 ;
			g_led_display[ ledindex ] = g_led_display[ ledindex ] + delta ;
		}
		if ( i > t ){
			uint8_t delta =  i - t ;
			delta = delta >> 4 ;// divide by 8
			if (delta == 0) delta = 1 ;
			g_led_display[ ledindex ] = g_led_display[ ledindex ] - delta ;
		}
	}
}

void set_binary( uint8_t data, uint8_t level_off, uint8_t level_on ){
	uint8_t bitmask = 1 ;
	for( uint8_t i = 0 ; i < 8 ; i++ ){
		g_led_target[ i ] = (data & bitmask) ? level_on : level_off ;
		bitmask <<= 1 ;
	}
}

uint8_t get_binary( uint8_t threshold ){
	uint8_t result = 0 ;
	uint8_t bitmask = 1 ;
	for( uint8_t i = 0 ; i < 8 ; i++ ){
		if (g_led_target[ i ] >= threshold) result |= bitmask ;
		bitmask <<= 1 ;
	}
	return result ;
}

#define is_within_visible_window( pos )((pos>=g_myfirstled)&&(pos<(g_myfirstled+8)))

void  __attribute__ ((noinline)) set_led( uint8_t pos, uint8_t value ){
	if ( is_within_visible_window(pos) ) g_led_target[ pos - g_myfirstled ] = value ;
}

void _decode_command( uint16_t cmd ){
	uint8_t operator ;
	uint8_t op1 ;
	uint8_t op2 ;
	operator = cmd >> 8 ;
	operator = operator >> 3 ;
	op1 = cmd >> 8 ;
	op1 = op1 & 0b111 ;
	op2 = cmd & 0b11111111 ;
	
//	set_binary( operator, 0, 255 ) ;
	
	switch( operator ){
		case CMD_ClearAll :
			set_binary( 0x00, 0 , 255 ) ;
			break ;
		case CMD_SetAll :
			set_binary( 0xff, 0 , 255 ) ;
			break ;
		case CMD_SetLED :
			set_led( op2, 255 ) ;
			break ;
		case CMD_ClearLED :
			set_led( op2, 0 ) ;
			break ;
		case CMD_SetBinary :
			if (op1 == (g_chain_pos - 1)){
				set_binary( op2, 0, 255 ) ;
			}
			break ;
		case CMD_Style :
			led_seteffect( op2 ) ;
			break ;
		case CMD_UserPattern :
			{
				uint8_t *e = eeprom_get_sweepuserpattern() ;
				uint8_t pattern = eeprom_read_byte( e + 2 + op2 ) ;
				set_binary( pattern, 0, 255 ) ;
			}
			break ;
		default :
			// do nothing
			break ;
	}
}
