/*
SweeperMeter

(C) Copyright, Nigel Batten, 2011
Contactable at <firstname>@batsocks.co.uk

This project is released under the
Creative Commons Attribution-Share Alike 3.0 licence.

See http://creativecommons.org/licenses/by-sa/3.0/

-----------------------------------------------------
An LED Sweeper with additional 'Meter' function.

8 LEDs, LED1 to LED8.
LED1 + LED2 are connected to PB4
LED3 + LED4 are connected to PB3
LED5 + LED6 are connected to PB1
LED7 + LED8 are connected to PB0
If a pin is "output 1", the odd-numbered LED is lit (L1,L3,L5,L7).
If a pin is "output 0", then even-numbered LED is lit (L2,L4,L6,L8).
If a pin is 'input' (and no pullup resistor), then neither LED is lit.

PB2 is the 'sense/data' pin.

*/
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "globals.h"
#include "led.h"
#include "reset.h"
#include "sweep.h"
#include "meter.h"
#include "display.h"
#include "setup.h"

function_handler * PROGMEM f_function_lookup[] =
{
	fn_showbyte,		//	FUNCTION_ERROR	= 0
	fn_sweep,				//	FUNCTION_SWEEP,
	fn_display,			//	FUNCTION_DISPLAY,
	fn_meter,				//	FUNCTION_METER,
	fn_thermometer,	//	FUNCTION_THERMOMETER,
	(void*) 0,			//	FUNCTION_POV (not yet written),
	(void*) 0,			//	FUNCTION_ROULETTE (not yet written),
	
	fn_menu_main,					//	FUNCTION_MENU_MAIN
	fn_menu_function,			//	FUNCTION_MENU_FUNCTION
	fn_menu_sweeper,			//	FUNCTION_MENU_SWEEPER
	fn_menu_meter,				//	FUNCTION_MENU_METER
	fn_menu_chain,				//	FUNCTION_MENU_CHAIN
	fn_menu_chain_length,	//	FUNCTION_MENU_CHAIN_LENGTH
	fn_menu_chain_pos,		//	FUNCTION_MENU_CHAIN_POS
	fn_menu_utils,				//	FUNCTION_MENU_UTILS
	fn_menu_tick,					//	FUNCTION_MENU_TICK
	fn_menu_transmit,			//	FUNCTION_MENU_TRANSMIT
	
	fn_setfunction,			//	FUNCTION_SET_FUNCTION
	fn_setmeter,				//	FUNCTION_SET_METER
	fn_setchainlength,	//	FUNCTION_SET_CHAIN_LENGTH
	fn_setchainpos,			//	FUNCTION_SET_CHAIN_POS
	fn_import,					//	FUNCTION_IMPORT
	fn_settick,					//	FUNCTION_SET_TICK
	fn_settransmit,			//	FUNCTION_SET_TRANSMIT
	fn_version					//	FUNCTION_VERSION
} ;

int __attribute__((OS_main)) main(void)
{
	reset_init(); // find the reset cause...

	switch( g_reset_reason ){
		case (1<<PORF) : // normal power-up reset
			eeprom_restore_state() ; // retrieve globals from eeprom
			// now deal with nonsense data.
			if (g_function_id == 0xff) g_function_id = FUNCTION_MENU_MAIN ;
			if (g_function_data == 0xff) g_function_data = 1 ;
			break ;
		case (1<<PORF)|(1<<EXTRF) : // power-up with button pressed reset
			// go into the setup menus...
			g_function_id = FUNCTION_MENU_MAIN ;
			g_function_data = 0 ;
			break ;
		case (1<<EXTRF) : // normal button press reset
			// all the variables are set up already from before the reset.
			// just run the previously chosen functions.
			break ;
		default :
			// an unknown reset cause...
			// could be brown-out + external reset (seen after reprogramming)
			g_function_id = FUNCTION_ERROR ;
			g_function_data = g_reset_reason ;
			break ;
	}

	led_init() ;
	
	if (g_function_id >= FUNCTION_MARKER_END) g_function_id = FUNCTION_ERROR ;
	function_handler *p_fn = (function_handler *)pgm_read_word( &f_function_lookup[ g_function_id ] ) ;
	if (p_fn) p_fn( g_function_data ) ;
	
	// the function wasn't found, or wasn't defined.
	// Just shown the binary value of g_function_id.
	fn_showbyte( g_function_id ) ;
}


