/* setup.c
*/

#include "setup.h"
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "led.h"
#include "globals.h"
#include "meter.h"
#include "sweep.h"
//#include "pov.h"
//#include "reset.h"

typedef uint8_t bool ;
#define false 0
#define true 1

uint8_t bit_reverse( uint8_t x )
{// swap the endian-ness of a byte.
	// stolen from this thread:
	// http://www.avrfreaks.net/index.php?name=PNphpBB2&file=viewtopic&t=65645&view=next
	// courtesy of C Lawson.
    x = ((x >> 1) & 0x55) | ((x << 1) & 0xaa);
    x = ((x >> 2) & 0x33) | ((x << 2) & 0xcc);
    x = ((x >> 4) & 0x0f) | ((x << 4) & 0xf0);
    return x;   
} 

// a single cycle of sine. 32 entries. 128 offset. 0 min. 255 max.
// starts at value 127 going up.
uint8_t f_sine[32] PROGMEM = {
	0x7F,0x97,0xAF,0xC5,0xD8,0xE8,0xF4,0xFB,
	0xFE,0xFB,0xF4,0xE8,0xD8,0xC5,0xAF,0x97,
	0x7F,0x66,0x4E,0x38,0x25,0x15,0x09,0x02,
	0x00,0x02,0x09,0x15,0x25,0x38,0x4E,0x66
} ;


void initmenupattern( void ){
	led_init() ;
	led_seteffect( STYLE_Direct ) ;
	sei();
}

void showpattern( uint8_t fixedpattern, uint8_t pulsingpattern, uint8_t blinkingpattern, uint8_t cycles ){
static uint8_t s_phase = 0 ;
uint8_t t[8] ;
	// set up the fixed bits...
	uint8_t bitmask = 1 ;
	for( uint8_t ledindex = 0 ; ledindex < 8 ; ledindex++ ){
		if (!(bitmask & (pulsingpattern | blinkingpattern))){
			t[ ledindex ] = (fixedpattern & bitmask) ? 255 : 0 ;
		}
		bitmask <<= 1 ;
	}

	// and now do the pulsed and blinking LEDs
	for( uint8_t count = 0 ; count < cycles ; count++ ){
		
		for( uint8_t loop = 0 ; loop < 10 ; loop++ ){
			s_phase++ ; // overflow to 0 is expected.
			bitmask = 1 ;
			uint8_t amp_slowphase = pgm_read_byte( f_sine + (s_phase & 0b11111)) ;
			uint8_t amp_fastphase = pgm_read_byte( f_sine + ((s_phase << 1) & 0b11111) ) ;
			for( uint8_t ledindex = 0 ; ledindex < 8 ; ledindex++ ){
				if (pulsingpattern & bitmask){
					t[ ledindex ] = 0b10000000 + (amp_slowphase >> 1) ;
				}
				if (blinkingpattern & bitmask){
					t[ ledindex ] = (amp_fastphase & 0b10000000) ? 255 : 0 ;
				}
				bitmask <<= 1 ;
			}
			cli();
			for( uint8_t i = 0 ; i < 8 ; i++ ){
				g_led_target[i] = t[i] ;
			}
			sei();
			_delay_ms( 10 ) ;
		}
	}
}

// display a pattern on the LEDs for a certain number of cycles.
void showsetuppattern(
	uint8_t option, // option to display. 0-15. Shown on LEDs 1-4 (steady)
	uint8_t pagenum,// pageID to display. 0-15. Shown reversed on LEDs 5-8 (blinking)
	bool pulse, // if set, the option LEDs will be pulsed.
	uint8_t cycles ) // the number of cycles to show for. Approx 10 cycles per second.
{
	uint8_t blinkingpattern = bit_reverse( pagenum & 0xf ) ;
	uint8_t fixedpattern = (option & 0xf) ;
	uint8_t pulsingpattern = (pulse)?fixedpattern:0 ;
	
	showpattern( fixedpattern, pulsingpattern, blinkingpattern, cycles ) ;
}


void runsetuppattern(
		uint8_t pageID,					// page ID. Shown on LEDs 5-8
		uint8_t option,					// option ID. Shown on LEDs 1-4
		FunctionID_t functionid,// function to call when chosen.
		uint8_t data,						// byte to pass to function.
		uint8_t *p_current_data // pointer to current setting. NULL for n/a
){
bool pulse = false ;
	eeprom_restore_state() ;

	if (p_current_data){
		if (data == *p_current_data) pulse = true ;
	}

	g_function_data = data ;
	g_function_id = functionid ;
	showsetuppattern( option, pageID, pulse, 30 ) ;
	showsetuppattern( 0,0,false,1 ) ;
}


typedef struct {
	FunctionID_t function_id ;
	uint8_t function_data ;
	uint8_t *p_global ; // pointer to global data variable that holds the current value
} t_menuoptiondata ;

void __attribute__((noreturn)) fn_menu_generic( uint8_t pageid, t_menuoptiondata *f_optiondata, uint8_t optioncount ){
t_menuoptiondata optiondata ;

	initmenupattern() ;
	while(1){
		for( uint8_t option = 0 ; option < optioncount ; option++ ){
			memcpy_P( &optiondata, &f_optiondata[option],sizeof(t_menuoptiondata) ) ;
			runsetuppattern( pageid, option + 1, optiondata.function_id, optiondata.function_data, optiondata.p_global ) ;
		}
	}
}

t_menuoptiondata f_menu_main[] PROGMEM = {
	{FUNCTION_MENU_FUNCTION,0,0},
	{FUNCTION_MENU_SWEEPER,0,0},
	{FUNCTION_MENU_METER,0,0},
	{FUNCTION_MENU_CHAIN,0,0},
	{FUNCTION_MENU_UTILS,0,0}
} ;

void __attribute__((noreturn)) fn_menu_main( uint8_t data ){
	fn_menu_generic( 1, f_menu_main, sizeof(f_menu_main)/sizeof(f_menu_main[0]) ) ;
}

t_menuoptiondata f_menu_sweeper[] PROGMEM = {
	{FUNCTION_MENU_MAIN,0,0},
	{FUNCTION_MENU_TICK,0,0},
	{FUNCTION_MENU_TRANSMIT,0,0},
} ;
void __attribute__((noreturn)) fn_menu_sweeper( uint8_t data ){
	fn_menu_generic( 3, f_menu_sweeper, sizeof(f_menu_sweeper)/sizeof(f_menu_sweeper[0]) ) ;
}

t_menuoptiondata f_menu_tick[] PROGMEM = {
	{FUNCTION_MENU_SWEEPER,0,0},
	{FUNCTION_SET_TICK,TICK_VERYSLOW,&g_tickdelay},
	{FUNCTION_SET_TICK,TICK_SLOW,&g_tickdelay},
	{FUNCTION_SET_TICK,TICK_MEDIUM,&g_tickdelay},
	{FUNCTION_SET_TICK,TICK_FAST,&g_tickdelay},
	{FUNCTION_SET_TICK,TICK_VERYFAST,&g_tickdelay},
} ;

void __attribute__((noreturn)) fn_menu_tick( uint8_t data ){
	fn_menu_generic( 5, f_menu_tick, sizeof(f_menu_tick)/sizeof(f_menu_tick[0]) ) ;
}

t_menuoptiondata f_menu_transmit[] PROGMEM = {
	{FUNCTION_MENU_SWEEPER,0,0},
	{FUNCTION_SET_TRANSMIT,TRANSMIT_DISABLE,&g_transmit},
	{FUNCTION_SET_TRANSMIT,TRANSMIT_ENABLE,&g_transmit},
} ;

void __attribute__((noreturn)) fn_menu_transmit( uint8_t data ){
	fn_menu_generic( 6, f_menu_transmit, sizeof(f_menu_transmit)/sizeof(f_menu_transmit[0]) ) ;
}

t_menuoptiondata f_menu_function[] PROGMEM = {
	{FUNCTION_MENU_MAIN,0,0},
	{FUNCTION_SET_FUNCTION,FUNCTION_SWEEP,&g_function_id},
	{FUNCTION_SET_FUNCTION,FUNCTION_METER,&g_function_id},
	{FUNCTION_SET_FUNCTION,FUNCTION_DISPLAY,&g_function_id},
	{FUNCTION_SET_FUNCTION,FUNCTION_ROULETTE,&g_function_id},
} ;

void __attribute__((noreturn)) fn_menu_function( uint8_t data ){
	fn_menu_generic( 2, f_menu_function, sizeof(f_menu_function)/sizeof(f_menu_function[0]) ) ;
}

t_menuoptiondata f_menu_meter[] PROGMEM = {
	{FUNCTION_MENU_MAIN,0,0},
	{FUNCTION_SET_METER,RANGE_RESET,0},
	{FUNCTION_SET_METER,RANGE_SETMIN,0},
	{FUNCTION_SET_METER,RANGE_SETMAX,0}
} ;

void __attribute__((noreturn)) fn_menu_meter( uint8_t data ){
	// start the ADC running...
	meter_init( SOURCE_SensePin ) ;
	fn_menu_generic( 4, f_menu_meter, sizeof(f_menu_meter)/sizeof(f_menu_meter[0]) ) ;
}

t_menuoptiondata f_menu_chain[] PROGMEM = {
	{FUNCTION_MENU_MAIN,0,0},
	{FUNCTION_MENU_CHAIN_LENGTH,0,0},
	{FUNCTION_MENU_CHAIN_POS,0,0}
} ;

void __attribute__((noreturn)) fn_menu_chain( uint8_t data ){
	uint8_t length = sizeof(f_menu_chain)/sizeof(f_menu_chain[0]) ;
	if (g_chain_length <= 1) length-- ; // don't display the Position menu.
	fn_menu_generic( 8, f_menu_chain, length ) ;
}

t_menuoptiondata f_menu_chain_length[] PROGMEM = {
	{FUNCTION_MENU_CHAIN,0,0},
	{FUNCTION_SET_CHAIN_LENGTH, 1, &g_chain_length},
	{FUNCTION_SET_CHAIN_LENGTH, 2, &g_chain_length},
	{FUNCTION_SET_CHAIN_LENGTH, 3, &g_chain_length},
	{FUNCTION_SET_CHAIN_LENGTH, 4, &g_chain_length},
} ;

void __attribute__((noreturn)) fn_menu_chain_length( uint8_t data ){
	fn_menu_generic( 9, f_menu_chain_length, sizeof(f_menu_chain_length)/sizeof(f_menu_chain_length[0]) ) ;
}

t_menuoptiondata f_menu_chain_pos[] PROGMEM = {
	{FUNCTION_MENU_CHAIN,0,0},
	{FUNCTION_SET_CHAIN_POS, 1, &g_chain_pos},
	{FUNCTION_SET_CHAIN_POS, 2, &g_chain_pos},
	{FUNCTION_SET_CHAIN_POS, 3, &g_chain_pos},
	{FUNCTION_SET_CHAIN_POS, 4, &g_chain_pos}
} ;
void __attribute__((noreturn)) fn_menu_chain_pos( uint8_t data ){
	// limit the options to the length of the chain!
	fn_menu_generic( 10, f_menu_chain_pos, g_chain_length + 1 ) ;
}

t_menuoptiondata f_menu_utils[] PROGMEM = {
	{FUNCTION_MENU_MAIN, 0, 0},
	{FUNCTION_VERSION, 0, 0},
	{FUNCTION_IMPORT, IMPORT_USER_PATTERN, 0 }
} ;
void __attribute__((noreturn)) fn_menu_utils( uint8_t data ){
	fn_menu_generic( 15, f_menu_utils, sizeof(f_menu_utils)/sizeof(f_menu_utils[0]) ) ;
}

void __attribute__((noreturn)) fn_showbyte( uint8_t data ){
	// firstly set up the initial state for if the user presses the button.
	// this will mean that the next press of the button will take the user
	// to their normal function.
	eeprom_restore_state() ;
	
	uint8_t bitmask = 1 ;
	for( uint8_t i = 0 ; i < 8 ; i++ ){ g_led_target[i] = (data & bitmask)?255:0 ; bitmask <<= 1 ;}
	sei();
	while(1){}
}

void __attribute((noreturn)) fn_version( uint8_t data ){
	g_function_id = FUNCTION_MENU_MAIN ;
	while(1){
		showpattern( 0, VERSION_MAJOR, 0, 30 ) ;
		showpattern( 0, 0, 0, 1 ) ;
		showpattern( VERSION_MINOR, 0, 0, 30 ) ;
		showpattern( 0, 0, 0, 1 ) ;
	}
}

void __attribute__((noreturn)) fn_setfunction( uint8_t data ){
	eeprom_restore_state() ;
	if (g_function_id == data){
		// the chosen function has not changed.
	} else {
		// the chosen function _has_ changed.
		// set the function and save the state...
		g_function_id = data ;
		g_function_data = 0 ; // initialise to zero
		eeprom_save_state();
	}
	// return to the main menu
	fn_menu_main(0) ;
}

void __attribute__((noreturn)) fn_setchainlength( uint8_t data ){
	eeprom_restore_state() ;
	if (g_chain_length != data){
		// the length of the chain has changed.
		// The default position in a chain is at the end.
		g_chain_length = data ;
		g_chain_pos = data ;
	}
	if (g_chain_length == 0) g_chain_length = 1 ;
	if (g_chain_length > 4) g_chain_length = 4 ;
	if (g_chain_pos > g_chain_length) g_chain_pos = g_chain_length ;
	g_myfirstled = (g_chain_pos - 1) * 8 ;
	g_chain_ledlength=  g_chain_length * 8 ;
	eeprom_save_state();
	
	// return to the chain menu
	fn_menu_chain( 0 ) ;
}

void __attribute__((noreturn)) fn_setchainpos( uint8_t data ){
	eeprom_restore_state() ;
	g_chain_pos = data ;
	if (g_chain_pos == 0) g_chain_pos = 1 ;
	if (g_chain_pos > g_chain_length) g_chain_pos = g_chain_length ;
	g_myfirstled = (g_chain_pos - 1) * 8 ;
	g_chain_ledlength=  g_chain_length * 8 ;
	eeprom_save_state();

	// return to the chain menu
	fn_menu_chain( 0 ) ;
}

void __attribute__((noreturn)) fn_setmeter( uint8_t data ){
	eeprom_restore_state() ;

	if (data == RANGE_RESET){
		g_meter_min = 0 ;
		g_meter_max = 1023 ;
	}

	if (data == RANGE_SETMIN){
		g_meter_min = meter_lastvalue() ;
	}

	if (data == RANGE_SETMAX){
		g_meter_max = meter_lastvalue() ;
	}

	eeprom_save_state() ;
	if (data == RANGE_RESET)
		fn_menu_main( 0 ) ;
	else
		fn_menu_meter( 0 ) ;
}

void __attribute__((noreturn)) fn_import( uint8_t data ){
	if (data == IMPORT_USER_PATTERN) fn_sweep_importpattern() ;
	fn_menu_main( 0 ) ;
}

void __attribute__((noreturn)) fn_settransmit( uint8_t data ){
	eeprom_restore_state() ;
	g_transmit = data ;
	eeprom_save_state() ;
	fn_menu_sweeper( 0 ) ;
}

void __attribute__((noreturn)) fn_settick( uint8_t data ){
	eeprom_restore_state() ;
	g_tickdelay = data ;
	eeprom_save_state() ;
	fn_menu_sweeper( 0 ) ;
}
